package de.hft.stuttgart.citydoctor2.healer.genetic;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;

public class GeneticAlgorithm<T> implements Runnable {

	private int populationSize = 5;
	private int maxGenerations = 2;

	private Crossover<T> cross;
	private Mutation<T> mutate;
	private Heuristic<T> heuristic;
	private FitnessEvaluation<T> evaluation;
	
	private Individual<T> result;

	private GeneticAlgorithm(Crossover<T> cross, Mutation<T> mutate, Heuristic<T> heuristic,
			FitnessEvaluation<T> evaluation) {
		this.cross = Objects.requireNonNull(cross);
		this.mutate = Objects.requireNonNull(mutate);
		this.heuristic = Objects.requireNonNull(heuristic);
		this.evaluation = Objects.requireNonNull(evaluation);
	}

	public static <T> GeneticAlgorithm<T> of(Crossover<T> cross, Mutation<T> mutate, Heuristic<T> heuristic,
			FitnessEvaluation<T> evaluation) {
		return new GeneticAlgorithm<>(cross, mutate, heuristic, evaluation);
	}

	@Override
	public void run() {
		// initial population
		List<Individual<T>> population = new ArrayList<>();
		for (int i = 0; i < populationSize; i++) {
			population.add(new Individual<>(heuristic.generate()));
		}
		for (Individual<T> genome : population) {
			genome.evaluateFitness(evaluation);
		}
		Collections.sort(population);
		System.out.println("Best Fitness: " + population.get(0).getFitness());
		// do the generations thing
		for (int i = 0; i < maxGenerations; i++) {
			// generate children
			for (int j = 0; j < populationSize - 1; j++) {
				// maybe random children generation
				Individual<T> parent1 = population.get(j);
				Individual<T> parent2 = population.get(j + 1);
				Individual<T> child1 = new Individual<>(cross.cross(parent1.getObject(), parent2.getObject()));
				Individual<T> child2 = new Individual<>(cross.cross(parent2.getObject(), parent1.getObject()));
				// mutate
				child1 = new Individual<>(mutate.mutate(child1.getObject()));
				child2 = new Individual<>(mutate.mutate(child2.getObject()));
				child1.evaluateFitness(evaluation);
				child2.evaluateFitness(evaluation);
				population.add(child1);
				population.add(child2);
			}
			// sort
			Collections.sort(population);
			// eliminate
			List<Individual<T>> newGeneration = new ArrayList<>(populationSize);
			for (int j = 0; j < populationSize; j++) {
				newGeneration.add(population.get(j));
			}
			population = newGeneration;
			System.out.println("Best Fitness: " + population.get(0).getFitness());
		}
		result = population.get(0);
		System.out.println("End. Best fitness: " + population.get(0).getFitness() + " with plan: " + population.get(0).getObject());
	}
	
	public Individual<T> getResult() {
		return result;
	}

}
